/*     
 **********************************************************************
 *     osutils.c - OS Services layer for emu10k1 driver 
 *     Copyright 1999, 2000 Creative Labs, Inc. 
 * 
 ********************************************************************** 
 * 
 *     Date                 Author          Summary of changes 
 *     ----                 ------          ------------------ 
 *     October 20, 1999     Bertrand Lee    base code release 
 *     November 2, 1999     Alan Cox        cleaned up
 * 
 ********************************************************************** 
 * 
 *     This program is free software; you can redistribute it and/or 
 *     modify it under the terms of the GNU General Public License as 
 *     published by the Free Software Foundation; either version 2 of 
 *     the License, or (at your option) any later version. 
 * 
 *     This program is distributed in the hope that it will be useful, 
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 *     GNU General Public License for more details. 
 * 
 *     You should have received a copy of the GNU General Public 
 *     License along with this program; if not, write to the Free 
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, 
 *     USA. 
 * 
 ********************************************************************** 
 */ 

#include <platform.h>

#include <linux/delay.h>
#include <linux/interrupt.h>
#include "hwaccess.h"
#include "mycommon.h"

extern spinlock_t sblive_spinlock;

int
osAllocMemPhysical (
    u32     size,
    u32 *    phandle,
    void **    linear,
    unsigned long *physical
    )
{
  {
    u32     reqpage, order;
    PLMEMHANDLE memhandle;

    memhandle = kmalloc(sizeof(LMEMHANDLE), GFP_KERNEL);
    if ( memhandle == NULL ) return CTSTATUS_NOMEMORY;
#ifdef MEMTEST
    PDEBUG("\nosutils.c: kmalloc: [%p]\n", memhandle);
#endif
    order = 0;
    reqpage = size/PAGE_SIZE;
    if ( size % PAGE_SIZE )
      reqpage++;
    if ( reqpage != 0 ) {
      reqpage--;
      while ( reqpage > 0 ) {
      	reqpage >>= 1;
	order++;
      }
    }
    DPD(" order is ", order);
#if LINUX_VERSION_CODE < 0x020100
    *linear = (void *)__get_free_pages(GFP_KERNEL, order, 0);
#else
    *linear = (void *)__get_free_pages(GFP_KERNEL, order);
#endif

    if (*linear == 0) 
      {
        kfree(memhandle);
	#ifdef MEMTEST
	  PDEBUG("\nosutils.c: kfree: [%p]\n", memhandle);
	#endif
	return CTSTATUS_NOMEMORY;
      }
    #ifdef MEMTEST
      PDEBUG("\nosutils.c: __get_free_pages: [%p]\n", *linear);
    #endif
    // in linux, we can directly access physical address, don't need to do 
    // phys_to_virt.
    // In linux kernel 2.0.36, virt_to_bus does nothing, get_free_pages
    // returns physical address. But in kernel 2.2.1 upwards, 
    // get_free_pages returns virtual address, we need to convert it
    // to physical address. Then this physical address can be used to program
    // hardware registers.
    *physical = virt_to_bus(*linear);
    memhandle->physaddx = *physical;
    memhandle->virtaddx = *linear;
    memhandle->order    = order;
    *phandle = (u32)memhandle;
    return CTSTATUS_SUCCESS;  
  }
}


int
osCopyPageTable (
    char    *virtaddr,
    u32     startPage,
    u32     numpages,
    u32     *paddr
    )
{

  u32 i;
  u32 physaddx;

  for (i = 0; i < numpages; i++, paddr++) {
	physaddx = (u32) virt_to_bus((void *) virtaddr);
	virtaddr = virtaddr + 4096;
	*paddr = physaddx;
  } 

    return CTSTATUS_SUCCESS;
}


int
osFreeMemPhysical(u32 handle)
{
  {
    PLMEMHANDLE memhandle = (PLMEMHANDLE)handle;
    if (!handle) return CTSTATUS_ERROR;
    free_pages((unsigned long)memhandle->virtaddx, memhandle->order);
    kfree(memhandle);
#ifdef MEMTEST
    PDEBUG("\nosutils.c: free_pages: [%p]\n", memhandle->virtaddx);
    PDEBUG("\nosutils.c: kfree: [%p]\n", memhandle);
#endif
    return CTSTATUS_SUCCESS;
  }

}


int osListAttach (struct sblive_list ** head, struct sblive_list * new)
{
    struct sblive_list * pTemp;
    unsigned long flags;

    if (head == NULL)
        return CTSTATUS_ERROR;

    spin_lock_irqsave(&sblive_spinlock, flags);

    new->next = NULL;
    if (*head == NULL)
    {
        *head = new;
    }
    else
    {
        pTemp = *head;
        while (pTemp->next != NULL) pTemp = pTemp->next;
        pTemp->next = new;
    }

    spin_unlock_irqrestore(&sblive_spinlock, flags);

    return CTSTATUS_SUCCESS;
}


int
osListRemove (struct sblive_list ** head, struct sblive_list * pDead)
{
    struct sblive_list * pTemp;
    u32 status = CTSTATUS_SUCCESS;
    unsigned long flags;

    if (head == NULL)
        return CTSTATUS_ERROR;

    spin_lock_irqsave(&sblive_spinlock, flags);

    if (*head == pDead)
    {
        *head = pDead->next;
    }
    else
    {
        pTemp = *head;
        while (pTemp != NULL && pTemp->next != pDead)
            pTemp = pTemp->next;

        if (pTemp == NULL)
            status = CTSTATUS_ERROR;
        else
        {
            pTemp->next = pDead->next;
            pDead->next = NULL;
        }
    }

    spin_unlock_irqrestore(&sblive_spinlock, flags);

    return status;
}


struct sblive_list *
osListGetNext(struct sblive_list * head, struct sblive_list * curr)
{
    if (curr == NULL)
        return head;
    return curr->next;
}

// DPC Handling

typedef struct {
    struct tq_struct *pTask;
    CALLBACKFN callback;
    u32 refdata;
    u32 param;
    int is_active;
} stDPCCbkHandle;

void task_queue_callback(void *pData)
{
    stDPCCbkHandle *phDPCCbk = pData;

    phDPCCbk->callback(phDPCCbk->refdata, phDPCCbk->param, 0);
    kfree(phDPCCbk->pTask);
    kfree(phDPCCbk);
}

int
osScheduleDPC (PCTDPC dpc)
{
    stDPCCbkHandle *phDPCCbk;
    struct tq_struct *pTask;

    pTask = kmalloc(sizeof(struct tq_struct), GFP_ATOMIC);
    phDPCCbk = kmalloc(sizeof(stDPCCbkHandle), GFP_ATOMIC);

    pTask->next = NULL;
    pTask->sync = 0;
    pTask->routine = task_queue_callback;
    pTask->data = phDPCCbk;

    phDPCCbk->pTask = pTask;
    phDPCCbk->callback = dpc->DPCCallBackFn;
    phDPCCbk->refdata = dpc->refdata;
    phDPCCbk->param = dpc->param;

    queue_task(pTask, &tq_immediate);
    mark_bh(IMMEDIATE_BH);

    return CTSTATUS_SUCCESS;
}


